#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
#
#  --------------------------------------------------------------------------
#
from .GXUInt8 import GXUInt8

# BitString class is used with Bit strings.
class GXBitString:

    #
    # Constructor.
    #
    # @param val
    # Bit string value.
    #
    def __init__(self, val, count=None):
        if isinstance(val, (GXUInt8)):
            val = GXBitString.___toBitString(val, 8)
            if count:
                val = val[0:count]
        self.value = val

    @classmethod
    def ___toBitString(cls, value, count):
        count = min(count, 8)
        pos = 0
        sb = ""
        while pos != count:
            if (value & (1 << pos)) != 0:
                sb += "1"
            else:
                sb += "0"
            pos = 1 + pos
        return sb

    # Convert integer value to BitString.
    # value: Value to convert.
    # count: Amount of bits.
    # Returns: Bitstring.
    @classmethod
    def toBitString(cls, value, count):
        sb = cls.___toBitString(value & 0xFF, count)
        if count > 8:
            sb += cls.___toBitString((value >> 8) & 0xFF, count - 8)
            if count > 16:
                sb += cls.___toBitString((value >> 16) & 0xFF, count - 16)
                if count > 24:
                    sb += cls.___toBitString((value >> 24) & 0xFF, count - 24)
        if len(sb) > count:
            return sb[0, count]
        return sb

    def __str__(self):
        return self.value

    #
    # Bit string value as byte.
    #
    def toInteger(self):
        val = 0
        if self.value:
            pos = 0
            for it in self.value:
                if it == '1':
                    val |= (1 << pos)
                elif it != '0':
                    raise ValueError("Invalid parameter.")
                pos = 1 + pos
        return val



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from __future__ import print_function
from .enums import Priority, ServiceClass, InterfaceType, Authentication, Standard, HdlcFrameType, Conformance
from .enums.DateTimeSkips import DateTimeSkips
from .ConnectionState import ConnectionState
from .objects.GXDLMSObjectCollection import GXDLMSObjectCollection
from .GXHdlcSettings import GXHdlcSettings

# This class includes DLMS communication settings.
# pylint: disable=bad-option-value,too-many-public-methods,too-many-instance-attributes,old-style-class
class GXDLMSSettings:
    #
    # Server sender frame sequence starting number.
    #
    __SERVER_START_SENDER_FRAME_SEQUENCE = 0x1E

    #
    # Server receiver frame sequence starting number.
    #
    __SERVER_START_RECEIVER_FRAME_SEQUENCE = 0xFE

    #
    # Client sender frame sequence starting number.
    #
    __CLIENT_START_SENDER_FRAME_SEQUENCE = 0xFE

    #
    # Client receiver frame sequence starting number.
    #
    __CLIENT_START_RCEIVER_FRAME_SEQUENCE = 0xE

    #
    # DLMS version number.
    #
    __DLMS_VERSION = 6
    __MAX_RECEIVE_PDU_SIZE = 0xFFFF

    @classmethod
    def getInitialConformance(cls, useLN):
        if useLN:
            return Conformance.BLOCK_TRANSFER_WITH_ACTION | Conformance.BLOCK_TRANSFER_WITH_SET_OR_WRITE | Conformance.BLOCK_TRANSFER_WITH_GET_OR_READ | Conformance.SET | Conformance.SELECTIVE_ACCESS | Conformance.ACTION | Conformance.MULTIPLE_REFERENCES | Conformance.GET | Conformance.ACCESS
        return Conformance.INFORMATION_REPORT | Conformance.READ | Conformance.UN_CONFIRMED_WRITE | Conformance.WRITE | Conformance.PARAMETERIZED_ACCESS | Conformance.MULTIPLE_REFERENCES

    #
    # Constructor.
    #
    def __init__(self, isServer):
        self.customChallenges = False
        self.ctoSChallenge = None
        self.stoCChallenge = None
        self.sourceSystemTitle = None
        self.invokeId = 0x1
        self.longInvokeID = 0x1
        self.priority = Priority.HIGH
        self.serviceClass = ServiceClass.CONFIRMED
        self.clientAddress = 0
        self.serverAddress = 0
        self.pushClientAddress = 0
        self.serverAddressSize = 0
        self.__useLogicalNameReferencing = True
        self.interfaceType = InterfaceType.HDLC
        self.authentication = Authentication.NONE
        self.password = None
        self.kek = None
        self.count = 0
        self.index = 0
        self.targetEphemeralKey = None
        self.dlmsVersion = self.__DLMS_VERSION
        self.connected = ConnectionState.NONE
        self.allowAnonymousAccess = False
        self.maxPduSize = self.__MAX_RECEIVE_PDU_SIZE
        self.maxServerPDUSize = self.__MAX_RECEIVE_PDU_SIZE
        self.startingPacketIndex = 1
        # Gets current block index.
        self.blockIndex = 1
        self.cipher = None
        self.blockNumberAck = 0
        self.protocolVersion = None
        self.isServer = isServer
        self.objects = GXDLMSObjectCollection()
        self.hdlc = GXHdlcSettings()
        self.gateway = None
        self.proposedConformance = GXDLMSSettings.getInitialConformance(self.__useLogicalNameReferencing)
        self.receiverFrame = 0
        self.senderFrame = 0
        self.resetFrameSequence()
        self.gbtWindowSize = 1
        self.userId = -1
        #Quality of service.
        self.qualityOfService = 0
        self.useUtc2NormalTime = False
        self.increaseInvocationCounterForGMacAuthentication = False
        self.dateTimeSkips = DateTimeSkips.NONE
        self.standard = Standard.DLMS
        self.negotiatedConformance = Conformance.NONE
        self.invokeID = 0
        self.command = 0
        self.commandType = 0
        self.useCustomChallenge = False
        self.preEstablishedSystemTitle = None

    #
    # Client to Server challenge.
    #
    def getCtoSChallenge(self):
        return self.ctoSChallenge

    #
    # @param value
    # Client to Server challenge.
    #
    def setCtoSChallenge(self, value):
        if not self.customChallenges or self.ctoSChallenge is None:
            self.ctoSChallenge = value

    #
    # Server to Client challenge.
    #
    def getStoCChallenge(self):
        return self.stoCChallenge

    #
    # @param value
    # Server to Client challenge.
    #
    def setStoCChallenge(self, value):
        if not self.customChallenges or self.stoCChallenge is None:
            self.stoCChallenge = value

    #
    # Is connection accepted.
    #
    def acceptConnection(self):
        return self.connected != ConnectionState.NONE or self.allowAnonymousAccess or (self.cipher and self.cipher.sharedSecret)

    #
    # Reset frame sequence.
    #
    def resetFrameSequence(self):
        if self.isServer:
            self.senderFrame = self.__SERVER_START_SENDER_FRAME_SEQUENCE
            self.receiverFrame = self.__SERVER_START_RECEIVER_FRAME_SEQUENCE
        else:
            self.senderFrame = self.__CLIENT_START_SENDER_FRAME_SEQUENCE
            self.receiverFrame = self.__CLIENT_START_RCEIVER_FRAME_SEQUENCE

    #pylint: disable=too-many-return-statements
    def checkFrame(self, frame_, xml):
        #  If notify
        if frame_ == 0x13:
            return True

        #  If U frame.
        if (frame_ & HdlcFrameType.U_FRAME) == HdlcFrameType.U_FRAME:
            if frame_ == 0x93:
                isEcho = not self.isServer and frame_ == 0x93 and (self.senderFrame == 0x10 or self.senderFrame == 0xfe) and self.receiverFrame == 0xE
                self.resetFrameSequence()
                return not isEcho
            if frame_ == 0x73 and not self.isServer:
                return self.senderFrame == 0xFE and self.receiverFrame == 0xE
            return True
        #  If S -frame.
        if (frame_ & HdlcFrameType.S_FRAME) == HdlcFrameType.S_FRAME:
            #If echo.
            if frame_ == (self.senderFrame & 0xF1):
                return False
            self.receiverFrame = self.increaseReceiverSequence(self.receiverFrame)
            return True
        #  Handle I-frame.
        if (self.senderFrame & 0x1) == 0:
            expected = GXDLMSSettings.increaseReceiverSequence(GXDLMSSettings.increaseSendSequence(self.receiverFrame))
            if frame_ == expected:
                self.receiverFrame = frame_
                return True
            #If the final bit is not set.
            if frame_ == (expected & ~0x10) and self.hdlc.windowSizeRX != 1:
                self.receiverFrame = frame_
                return True
            # If Final bit is not set for the previous message.
            if (self.receiverFrame & 0x10) == 0 and self.hdlc.windowSizeRX != 1:
                expected = (0x10 | self.increaseSendSequence(self.receiverFrame))
                if frame_ == expected:
                    self.receiverFrame = frame_
                    return True
                # If the final bit is not set.
                if frame_ == (expected & ~0x10):
                    ReceiverFrame = frame_
                    return True
        else:
            expected = self.increaseSendSequence(self.receiverFrame) & 0xFF
            #  If answer for RR.
            if frame_ == expected:
                self.receiverFrame = frame_
                return True
            if frame_ == (expected & ~0x10):
                self.receiverFrame = frame_
                return True
            if self.hdlc.windowSizeRX != 1:
                #If HDLC window size is bigger than one.
                expected = self.increaseReceiverSequence(self.increaseSendSequence(ReceiverFrame))
                if frame_ == expected:
                    ReceiverFrame = frame_
                    return True

        # If try to find data from bytestream and not real communicating.
        if xml and ((not self.isServer and self.receiverFrame == 0xE) or\
            (self.isServer and self.receiverFrame == 0xEE)):
            ReceiverFrame = frame_
            return True
        print("Invalid HDLC Frame: " + hex(frame_) + " Expected: " + hex(expected))
        return False

    #
    # Increase receiver sequence.
    #
    # @param value
    # Frame value.
    # Increased receiver frame sequence.
    #
    @classmethod
    def increaseReceiverSequence(cls, value):
        return ((value & 0xFF) + 0x20 | 0x10 | value & 0xE) & 0xFF

    #
    # Increase sender sequence.
    #
    # @param value
    # Frame value.
    # Increased sender frame sequence.
    #
    @classmethod
    def increaseSendSequence(cls, value):
        return ((value & 0xF0 | (value + 0x2) & 0xE) & 0xFF) & 0xFF

    #
    # Generates I-frame.
    # @param first
    # Is this first packet.
    # Generated I-frame
    #
    def getNextSend(self, first):
        if first:
            self.senderFrame = self.increaseReceiverSequence(self.increaseSendSequence(int(self.senderFrame)))
        else:
            self.senderFrame = self.increaseSendSequence(int(self.senderFrame))
        return self.senderFrame & 0xFF

    #
    # Generates Receiver Ready S-frame.
    #
    def getReceiverReady(self):
        self.senderFrame = self.increaseReceiverSequence((self.senderFrame | 1))
        return self.senderFrame & 0xF1

    #
    # Generates Keep Alive S-frame.
    #
    def getKeepAlive(self):
        self.senderFrame = (self.senderFrame | 1)
        return self.senderFrame & 0xF1

    #
    # Gets starting block index in HDLC framing.  Default is One based,
    #      but some
    # meters use Zero based value.  Usually this is not used.
    #
    # Current block index.
    #
    def getStartingPacketIndex(self):
        return self.startingPacketIndex

    #
    # Set starting block index in HDLC framing.  Default is One based,
    #      but some
    # meters use Zero based value.  Usually this is not used.
    #
    # @param value
    # Zero based starting index.
    #
    def setStartingPacketIndex(self, value):
        self.startingPacketIndex = value
        self.resetBlockIndex()

    #
    # Sets current block index.
    #
    # @param value
    # Block index.
    #
    def setBlockIndex(self, value):
        self.blockIndex = value

    #
    # Block number acknowledged in GBT.
    #
    def getBlockNumberAck(self):
        return self.blockNumberAck

    #
    # @param value
    # Block number acknowledged in GBT.
    #
    def setBlockNumberAck(self, value):
        self.blockNumberAck = value

    #
    # Resets block index to default value.
    #
    def resetBlockIndex(self):
        self.blockIndex = self.startingPacketIndex
        self.blockNumberAck = 0

    #
    # Increases block index.
    #
    def increaseBlockIndex(self):
        self.blockIndex += 1

    # Is Logical Name Referencing used.
    # Don't use property. For some reason Python 2.7 doesn't call it in Rasbperry PI.
    def getUseLogicalNameReferencing(self):
        return self.__useLogicalNameReferencing

    # Is Logical Name Referencing used.
    # Don't use property. For some reason Python 2.7 doesn't call it in Rasbperry PI.
    def setUseLogicalNameReferencing(self, value):
        if self.__useLogicalNameReferencing != value:
            self.__useLogicalNameReferencing = value
            self.proposedConformance = GXDLMSSettings.getInitialConformance(self.__useLogicalNameReferencing)

    #
    # Invoke ID.
    #
    def getInvokeID(self):
        return self.invokeID

    #
    # @param value
    # update invoke ID.
    #
    def updateInvokeId(self, value):
        #pylint: disable=bad-option-value,redefined-variable-type
        if (value & 0x80) != 0:
            self.priority = Priority.HIGH
        else:
            self.priority = Priority.NORMAL
        if (value & 0x40) != 0:
            self.serviceClass = ServiceClass.CONFIRMED
        else:
            self.serviceClass = ServiceClass.UN_CONFIRMED
        self.invokeID = int((value & 0xF))

    #
    # @param value
    # Invoke ID.
    #
    def setInvokeID(self, value):
        if value > 0xF:
            raise ValueError("Invalid InvokeID")
        self.invokeID = int(value)

    #
    # Invoke ID.
    #
    def getLongInvokeID(self):
        return self.longInvokeID

    #
    # @param value
    # Invoke ID.
    #
    def setLongInvokeID(self, value):
        if value > 0xFFFFFFFF:
            raise ValueError("Invalid InvokeID")
        self.longInvokeID = value

    #
    # Source system title.
    #
    def getSourceSystemTitle(self):
        return self.sourceSystemTitle

    #
    # @param value
    # Source system title.
    #
    def setSourceSystemTitle(self, value):
        if not value or len(value) != 8:
            raise ValueError("Invalid client system title.")
        self.sourceSystemTitle = value

    #
    # Long data count.
    #
    def getCount(self):
        return self.count

    #
    # @param value
    # Data count.
    #
    def setCount(self, value):
        if value < 0:
            raise ValueError("Invalid count.")
        self.count = value

    #
    # Long data index.
    #
    def getIndex(self):
        return self.index

    #
    # @param value
    # Long data index
    #
    def setIndex(self, value):
        if value < 0:
            raise ValueError("Invalid Index.")
        self.index = value



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright =c Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from ..GXIntFlag import GXIntFlag

class DateTimeSkips(GXIntFlag):
    """Enumerated date time skipped fields."""
    #pylint: disable=too-few-public-methods

     # Nothing is skipped from date time.
    NONE = 0

     # Year part of date time is skipped.
    YEAR = 1

     # Month part of date time is skipped.
    MONTH = 2

     # Day part is skipped.
    DAY = 4

     # Day of week part of date time is skipped.
    DAY_OF_WEEK = 8

     # Hours part of date time is skipped.
    HOUR = 0x10

     # Minute part of date time is skipped.
    MINUTE = 0x20

     # Second part of date time is skipped.
    SECOND = 0x40

     # Hundreds of seconds part of date time is skipped.
    MILLISECOND = 0x80

     # Devitation is not used.
    DEVITATION = 0x100

     # Status is not used.
    STATUS = 0x200



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------

from .enums.Command import Command
from .GetCommandType import GetCommandType
#
#  LN Parameters
#pylint: disable=too-many-instance-attributes, too-many-arguments
class GXDLMSLNParameters:
    #
    # Constructor.
    #
    # forSettings: DLMS settings.
    # forInvokeId: Invoke ID.
    # forCommand: Command.
    # forCommandType: Command type.
    # forAttributeDescriptor: Attribute descriptor.
    # forData: Data.
    def __init__(self, forSettings, forInvokeId, forCommand, forCommandType, forAttributeDescriptor, forData, forStatus):
        # DLMS settings.
        self.settings = forSettings
        self.invokeId = forInvokeId
        self.blockIndex = self.settings.blockIndex
        self.blockNumberAck = self.settings.blockNumberAck
        # DLMS Command.
        self.command = forCommand
        # Request type.
        self.requestType = forCommandType
        # Attribute descriptor.
        self.attributeDescriptor = forAttributeDescriptor
        # Data.
        self.data = forData
        # Send date and time.  This is used in Data notification messages.
        self.time = None
        # Reply status.
        self.status = forStatus
        # Are there more data to send or more data to receive.
        self.multipleBlocks = forSettings.count != forSettings.index
        # Is this last block in send.
        self.lastBlock = forSettings.count == forSettings.index
        self.gbtWindowSize = 1
        # Is GBT streaming used.
        self.streaming = False
        if self.settings:
            self.settings.command = forCommand
            if forCommand == Command.GET_REQUEST and forCommandType != GetCommandType.NEXT_DATA_BLOCK:
                self.settings.commandType = forCommandType



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .GXDLMSObject import GXDLMSObject
from .IGXDLMSBase import IGXDLMSBase
from ..enums import ErrorCode
from ..internal._GXCommon import _GXCommon
from ..GXByteBuffer import GXByteBuffer
from ..enums import ObjectType, DataType
from .enums import ServiceType, MessageType
from .GXDLMSCaptureObject import GXDLMSCaptureObject

# pylint: disable=too-many-instance-attributes
class GXDLMSPushSetup(GXDLMSObject, IGXDLMSBase):
    """
    Online help:
    http://www.gurux.fi/Gurux.DLMS.Objects.GXDLMSPushSetup
    """

    def __init__(self, ln="0.7.25.9.0.255", sn=0):
        """
        Constructor.

        ln : Logical Name of the object.
        sn : Short Name of the object.
        """
        #pylint: disable=super-with-arguments
        super(GXDLMSPushSetup, self).__init__(ObjectType.PUSH_SETUP, ln, sn)
        self.pushObjectList = []
        self.communicationWindow = []
        self.service = ServiceType.TCP
        self.message = MessageType.COSEM_APDU
        self.destination = None
        self.randomisationStartInterval = 0
        self.numberOfRetries = 0
        self.repetitionDelay = 0

    def getValues(self):
        return [self.logicalName,
                self.pushObjectList,
                (self.service, self.destination, self.message),
                self.communicationWindow,
                self.randomisationStartInterval,
                self.numberOfRetries,
                self.repetitionDelay]

    def getPushValues(self, client, values):
        """
        Get received objects from push message.

        values: Received values.
        Returns clone of captured COSEM objects.
        """
        if len(values) != len(self.pushObjectList):
            raise ValueError("Size of the push object list is different than values.")
        pos = 0
        objects = []
        for k, v in self.pushObjectList:
            co = GXDLMSCaptureObject()
            co.attributeIndex = v.attributeIndex
            co.dataIndex = v.dataIndex
            objects.append((k, co))
            if v.attributeIndex == 0:
                tmp = values[pos]
                index = 1
                while index <= k.getAttributeCount():
                    client.updateValue(k, index, tmp[index - 1])
                    index = 1 + index
            else:
                client.updateValue(k, v.attributeIndex, values[pos])
            pos = 1 + pos

    def invoke(self, settings, e):
        if e.index != 1:
            e.error = ErrorCode.READ_WRITE_DENIED

    #
    # Activates the push process.
    #
    def activate(self, client):
        return client.method(self.getName(), self.objectType, 1, 0, DataType.INT8)

    def getAttributeIndexToRead(self, all_):
        attributes = list()
        #  LN is static and read only once.
        if all_ or not self.logicalName:
            attributes.append(1)
        #  PushObjectList
        if all_ or self.canRead(2):
            attributes.append(2)
        #  SendDestinationAndMethod
        if all_ or self.canRead(3):
            attributes.append(3)
        #  CommunicationWindow
        if all_ or self.canRead(4):
            attributes.append(4)
        #  RandomisationStartInterval
        if all_ or self.canRead(5):
            attributes.append(5)
        #  NumberOfRetries
        if all_ or self.canRead(6):
            attributes.append(6)
        #  RepetitionDelay
        if all_ or self.canRead(7):
            attributes.append(7)
        return attributes

    #
    # Returns amount of attributes.
    #
    def getAttributeCount(self):
        return 7

    #
    # Returns amount of methods.
    #
    def getMethodCount(self):
        return 1

    def getDataType(self, index):
        if index == 1:
            ret = DataType.OCTET_STRING
        elif index == 2:
            ret = DataType.ARRAY
        elif index == 3:
            ret = DataType.STRUCTURE
        elif index == 4:
            ret = DataType.ARRAY
        elif index == 5:
            ret = DataType.UINT16
        elif index == 6:
            ret = DataType.UINT8
        elif index == 7:
            ret = DataType.UINT16
        else:
            raise ValueError("getDataType failed. Invalid attribute index.")
        return ret
    #
    # Returns value of given attribute.
    #
    def getValue(self, settings, e):
        #pylint: disable=bad-option-value,redefined-variable-type
        buff = GXByteBuffer()
        if e.index == 1:
            ret = _GXCommon.logicalNameToBytes(self.logicalName)
        elif e.index == 2:
            buff.setUInt8(DataType.ARRAY)
            _GXCommon.setObjectCount(len(self.pushObjectList), buff)
            for k, v in self.pushObjectList:
                buff.setUInt8(DataType.STRUCTURE)
                buff.setUInt8(4)
                _GXCommon.setData(settings, buff, DataType.UINT16, k.objectType)
                _GXCommon.setData(settings, buff, DataType.OCTET_STRING, _GXCommon.logicalNameToBytes(k.logicalName))
                _GXCommon.setData(settings, buff, DataType.INT8, v.attributeIndex)
                _GXCommon.setData(settings, buff, DataType.UINT16, v.dataIndex)
            ret = buff
        elif e.index == 3:
            buff.setUInt8(DataType.STRUCTURE)
            buff.setUInt8(3)
            _GXCommon.setData(settings, buff, DataType.ENUM, self.service)
            if self.destination:
                _GXCommon.setData(settings, buff, DataType.OCTET_STRING, self.destination.encode())
            else:
                _GXCommon.setData(settings, buff, DataType.OCTET_STRING, None)
            _GXCommon.setData(settings, buff, DataType.ENUM, self.message)
            ret = buff
        elif e.index == 4:
            buff.setUInt8(DataType.ARRAY)
            _GXCommon.setObjectCount(len(self.communicationWindow), buff)
            for k, v in self.communicationWindow:
                buff.setUInt8(DataType.STRUCTURE)
                buff.setUInt8(2)
                _GXCommon.setData(settings, buff, DataType.OCTET_STRING, k)
                _GXCommon.setData(settings, buff, DataType.OCTET_STRING, v)
            return buff
        elif e.index == 5:
            ret = self.randomisationStartInterval
        elif e.index == 6:
            ret = self.numberOfRetries
        elif e.index == 7:
            ret = self.repetitionDelay
        else:
            e.error = ErrorCode.READ_WRITE_DENIED
        return ret

    def setValue(self, settings, e):
        if e.index == 1:
            self.logicalName = _GXCommon.toLogicalName(e.value)
        elif e.index == 2:
            self.pushObjectList = []
            #pylint: disable=import-outside-toplevel
            from .._GXObjectFactory import _GXObjectFactory
            if e.value:
                for it in e.value:
                    type_ = it[0]
                    ln = _GXCommon.toLogicalName(it[1])
                    obj = settings.objects.findByLN(type_, ln)
                    if not obj:
                        obj = _GXObjectFactory.createObject(type_)
                        obj.logicalName = ln
                    co = GXDLMSCaptureObject()
                    co.attributeIndex = it[2]
                    co.dataIndex = it[3]
                    self.pushObjectList.append((obj, co))
        elif e.index == 3:
            #pylint: disable=broad-except
            if e.value:
                self.service = ServiceType(e.value[0])
                try:
                    if self.service == ServiceType.HDLC:
                        self.destination = _GXCommon.toLogicalName(e.value[1])
                    else:
                        self.destination = e.value[1].decode()
                        #If destination is not ASCII string.
                        if self.destination and not GXByteBuffer.isAsciiString(self.destination):
                            self.destination = GXByteBuffer.toHex(self.destination)
                except Exception:
                    self.destination = GXByteBuffer.hex(e.value[1])
                self.message = e.value[2]
        elif e.index == 4:
            self.communicationWindow = []
            if e.value:
                for it in e.value:
                    start = _GXCommon.changeType(settings, it[0], DataType.DATETIME)
                    end = _GXCommon.changeType(settings, it[1], DataType.DATETIME)
                    self.communicationWindow.append((start, end))
        elif e.index == 5:
            self.randomisationStartInterval = e.value
        elif e.index == 6:
            self.numberOfRetries = e.value
        elif e.index == 7:
            self.repetitionDelay = e.value
        else:
            e.error = ErrorCode.READ_WRITE_DENIED

    def load(self, reader):
        from .._GXObjectFactory import _GXObjectFactory
        self.pushObjectList = []
        if reader.isStartElement("ObjectList", True):
            while reader.isStartElement("Item", True):
                ot = ObjectType(reader.readElementContentAsInt("ObjectType"))
                ln = reader.readElementContentAsString("LN")
                ai = reader.readElementContentAsInt("AI")
                di = reader.readElementContentAsInt("DI")
                reader.readEndElement("ObjectList")
                co = GXDLMSCaptureObject(ai, di)
                obj = reader.objects.findByLN(ot, ln)
                if not obj:
                    obj = _GXObjectFactory.createObject(ot)
                    obj.LogicalName = ln
                self.pushObjectList.append((obj, co))
            reader.readEndElement("ObjectList")
        self.service = ServiceType(reader.readElementContentAsInt("Service"))
        self.destination = reader.readElementContentAsString("Destination")
        self.message = MessageType(reader.readElementContentAsInt("Message"))
        self.communicationWindow = []
        if reader.isStartElement("CommunicationWindow", True):
            while reader.isStartElement("Item", True):
                start = reader.readElementContentAsDateTime("Start")
                end = reader.readElementContentAsDateTime("End")
                self.communicationWindow.append((start, end))
            reader.readEndElement("CommunicationWindow")
        self.randomisationStartInterval = reader.readElementContentAsInt("RandomisationStartInterval")
        self.numberOfRetries = reader.readElementContentAsInt("NumberOfRetries")
        self.repetitionDelay = reader.readElementContentAsInt("RepetitionDelay")

    def save(self, writer):
        if self.pushObjectList:
            writer.writeStartElement("ObjectList")
            for k, v in self.pushObjectList:
                writer.writeStartElement("Item")
                writer.writeElementString("ObjectType", int(k.objectType))
                writer.writeElementString("LN", k.logicalName)
                writer.writeElementString("AI", v.attributeIndex)
                writer.writeElementString("DI", v.dataIndex)
                writer.writeEndElement()
            writer.writeEndElement()
        writer.writeElementString("Service", int(self.service))
        writer.writeElementString("Destination", self.destination)
        writer.writeElementString("Message", int(self.message))
        writer.writeStartElement("CommunicationWindow")
        if self.communicationWindow:
            for k, v in self.communicationWindow:
                writer.writeStartElement("Item")
                writer.writeElementString("Start", k)
                writer.writeElementString("End", v)
                writer.writeEndElement()
        writer.writeEndElement()
        writer.writeElementString("RandomisationStartInterval", self.randomisationStartInterval)
        writer.writeElementString("NumberOfRetries", self.numberOfRetries)
        writer.writeElementString("RepetitionDelay", self.repetitionDelay)



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .GXDLMSObject import GXDLMSObject
from .IGXDLMSBase import IGXDLMSBase
from ..enums import ErrorCode
from ..internal._GXCommon import _GXCommon
from ..enums import ObjectType, DataType
from ..GXByteBuffer import GXByteBuffer
from .GXMacMulticastEntry import GXMacMulticastEntry
from .GXMacDirectTable import GXMacDirectTable
from .GXMacAvailableSwitch import GXMacAvailableSwitch
from .GXMacPhyCommunication import GXMacPhyCommunication

# pylint: disable=too-many-instance-attributes
class GXDLMSPrimeNbOfdmPlcMacNetworkAdministrationData(GXDLMSObject, IGXDLMSBase):
    """
    Online help:
    http://www.gurux.fi/Gurux.DLMS.Objects.GXDLMSPrimeNbOfdmPlcMacNetworkAdministrationData
    """

    def __init__(self, ln="0.0.28.5.0.255", sn=0):
        """
        Constructor.

        ln : Logical Name of the object.
        sn : Short Name of the object.
        """
        GXDLMSObject.__init__(self, ObjectType.PRIME_NB_OFDM_PLC_MAC_NETWORK_ADMINISTRATION_DATA, ln, sn)
        # List of entries in multicast switching table.
        self.multicastEntries = list()
        # Switch table.
        self.switchTable = list()
        # List of entries in multicast switching table.
        self.directTable = list()
        # List of available switches.
        self.availableSwitches = list()
        # List of PHY communication parameters.
        self.communications = list()

    def getValues(self):
        return [self.logicalName,
                self.multicastEntries,
                self.switchTable,
                self.directTable,
                self.availableSwitches,
                self.communications]

    def reset(self, client):
        """Reset the values."""
        return client.method(self.getName(), self.objectType, 1, 0, DataType.INT8)

    def invoke(self, settings, e):
        #  Resets the value to the default value.
        #  The default value is an instance specific constant.
        if e.index == 1:
            self.multicastEntries = list()
            self.switchTable = list()
            self.directTable = list()
            self.availableSwitches = list()
            self.communications = list()
        else:
            e.error = ErrorCode.READ_WRITE_DENIED
    #
    # Returns collection of attributes to read.  If attribute is static and
    # already read or device is returned HW error it is not returned.
    #
    def getAttributeIndexToRead(self, all_):
        attributes = list()
        #  LN is static and read only once.
        if all_ or not self.logicalName:
            attributes.append(1)
        #  MulticastEntries
        if all_ or self.canRead(2):
            attributes.append(2)
        #  SwitchTable
        if all_ or self.canRead(3):
            attributes.append(3)
        #  DirectTable
        if all_ or self.canRead(4):
            attributes.append(4)
        #  AvailableSwitches
        if all_ or self.canRead(5):
            attributes.append(5)
        #  Communications
        if all_ or self.canRead(6):
            attributes.append(6)
        return attributes

    #
    # Returns amount of attributes.
    #
    def getAttributeCount(self):
        return 6

    #
    # Returns amount of methods.
    #
    def getMethodCount(self):
        return 1

    def getDataType(self, index):
        if index == 1:
            return DataType.OCTET_STRING
        if index in (2, 3, 4, 5, 6):
            return DataType.ARRAY
        raise ValueError("getDataType failed. Invalid attribute index.")

    def __getMulticastEntries(self, settings):
        bb = GXByteBuffer()
        bb.setUInt8(DataType.ARRAY)
        if not self.multicastEntries:
            _GXCommon.setObjectCount(0, bb)
        else:
            _GXCommon.setObjectCount(len(self.multicastEntries), bb)
            for it in self.multicastEntries:
                bb.setUInt8(DataType.STRUCTURE)
                bb.setUInt8(2)
                _GXCommon.setData(settings, bb, DataType.INT8, it.id)
                _GXCommon.setData(settings, bb, DataType.INT16, it.members)
        return bb.array()

    def __getSwitchTable(self, settings):
        bb = GXByteBuffer()
        bb.setUInt8(DataType.ARRAY)
        if not self.switchTable:
            _GXCommon.setObjectCount(0, bb)
        else:
            _GXCommon.setObjectCount(len(self.switchTable), bb)
            for it in self.switchTable:
                _GXCommon.setData(settings, bb, DataType.INT16, it)
        return bb.array()

    def __getDirectTable(self, settings):
        bb = GXByteBuffer()
        bb.setUInt8(DataType.ARRAY)
        if not self.directTable:
            _GXCommon.setObjectCount(0, bb)
        else:
            _GXCommon.setObjectCount(len(self.directTable), bb)
            for it in self.directTable:
                bb.setUInt8(DataType.STRUCTURE)
                bb.setUInt8(7)
                _GXCommon.setData(settings, bb, DataType.INT16, it.sourceSId)
                _GXCommon.setData(settings, bb, DataType.INT16, it.sourceLnId)
                _GXCommon.setData(settings, bb, DataType.INT16, it.sourceLcId)
                _GXCommon.setData(settings, bb, DataType.INT16, it.destinationSId)
                _GXCommon.setData(settings, bb, DataType.INT16, it.destinationLnId)
                _GXCommon.setData(settings, bb, DataType.INT16, it.destinationLcId)
                _GXCommon.setData(settings, bb, DataType.OCTET_STRING, it.did)
        return bb.array()

    def __getAvailableSwitches(self, settings):
        bb = GXByteBuffer()
        bb.setUInt8(DataType.ARRAY)
        if not self.availableSwitches:
            _GXCommon.setObjectCount(0, bb)
        else:
            _GXCommon.setObjectCount(len(self.availableSwitches), bb)
            for it in self.availableSwitches:
                bb.setUInt8(DataType.STRUCTURE)
                bb.setUInt8(5)
                _GXCommon.setData(settings, bb, DataType.OCTET_STRING, it.sna)
                _GXCommon.setData(settings, bb, DataType.INT16, it.lsId)
                _GXCommon.setData(settings, bb, DataType.INT8, it.level)
                _GXCommon.setData(settings, bb, DataType.INT8, it.rxLevel)
                _GXCommon.setData(settings, bb, DataType.INT8, it.rxSnr)
        return bb.array()

    def __getCommunications(self, settings):
        bb = GXByteBuffer()
        bb.setUInt8(DataType.ARRAY)
        if not self.communications:
            _GXCommon.setObjectCount(0, bb)
        else:
            _GXCommon.setObjectCount(len(self.communications), bb)
            for it in self.communications:
                bb.setUInt8(DataType.STRUCTURE)
                bb.setUInt8(9)
                _GXCommon.setData(settings, bb, DataType.OCTET_STRING, it.eui)
                _GXCommon.setData(settings, bb, DataType.INT8, it.txPower)
                _GXCommon.setData(settings, bb, DataType.INT8, it.txCoding)
                _GXCommon.setData(settings, bb, DataType.INT8, it.rxCoding)
                _GXCommon.setData(settings, bb, DataType.INT8, it.rxLvl)
                _GXCommon.setData(settings, bb, DataType.INT8, it.snr)
                _GXCommon.setData(settings, bb, DataType.INT8, it.txPowerModified)
                _GXCommon.setData(settings, bb, DataType.INT8, it.txCodingModified)
                _GXCommon.setData(settings, bb, DataType.INT8, it.rxCodingModified)
        return bb.array()

    #
    # Returns value of given attribute.
    #
    def getValue(self, settings, e):
        if e.index == 1:
            ret = _GXCommon.logicalNameToBytes(self.logicalName)
        elif e.index == 2:
            ret = self.__getMulticastEntries(settings)
        elif e.index == 3:
            ret = self.__getSwitchTable(settings)
        elif e.index == 4:
            ret = self.__getDirectTable(settings)
        elif e.index == 5:
            ret = self.__getAvailableSwitches(settings)
        elif e.index == 6:
            ret = self.__getCommunications(settings)
        else:
            ret = None
            e.error = ErrorCode.READ_WRITE_DENIED
        return ret

    @classmethod
    def __setMulticastEntry(cls, value):
        data = list()
        if value:
            for it in value:
                v = GXMacMulticastEntry()
                v.id = it[0]
                v.members = it[1]
                data.append(v)
        return data

    @classmethod
    def __setSwitchTable(cls, value):
        data = list()
        if value:
            for it in value:
                data.append(it)
        return data

    @classmethod
    def __setDirectTable(cls, value):
        data = list()
        if value:
            for it in value:
                v = GXMacDirectTable()
                v.sourceSId = it[0]
                v.sourceLnId = it[1]
                v.sourceLcId = it[2]
                v.destinationSId = it[3]
                v.destinationLnId = it[4]
                v.destinationLcId = it[5]
                v.did = it[6]
                data.append(v)
        return data

    @classmethod
    def __setAvailableSwitches(cls, value):
        data = list()
        if value:
            for it in value:
                v = GXMacAvailableSwitch()
                v.sna = it[0]
                v.lsId = it[1]
                v.level = it[2]
                v.rxLevel = it[3]
                v.rxSnr = it[4]
                data.append(v)
        return data

    @classmethod
    def __setCommunications(cls, value):
        data = list()
        if value:
            for it in value:
                v = GXMacPhyCommunication()
                v.eui = it[0]
                v.txPower = it[1]
                v.txCoding = it[2]
                v.rxCoding = it[3]
                v.rxLvl = it[4]
                v.snr = it[5]
                v.txPowerModified = it[6]
                v.txCodingModified = it[7]
                v.rxCodingModified = it[8]
                data.append(v)
        return data

    #
    # Set value of given attribute.
    #
    def setValue(self, settings, e):
        if e.index == 1:
            self.logicalName = _GXCommon.toLogicalName(e.value)
        elif e.index == 2:
            self.multicastEntries = self.__setMulticastEntry(e.value)
        elif e.index == 3:
            self.switchTable = self.__setSwitchTable(e.value)
        elif e.index == 4:
            self.directTable = self.__setDirectTable(e.value)
        elif e.index == 5:
            self.availableSwitches = self.__setAvailableSwitches(e.value)
        elif e.index == 6:
            self.communications = self.__setCommunications(e.value)
        else:
            e.error = ErrorCode.READ_WRITE_DENIED

    @classmethod
    def __loadMulticastEntries(cls, reader):
        list_ = list()
        if reader.isStartElement("MulticastEntries", True):
            while reader.isStartElement("Item", True):
                it = GXMacMulticastEntry()
                list_.append(it)
                it.id = reader.readElementContentAsInt("Id")
                it.members = reader.readElementContentAsInt("Members")
            reader.readEndElement("MulticastEntries")
        return list_

    @classmethod
    def __loadSwitchTable(cls, reader):
        list_ = list()
        if reader.isStartElement("SwitchTable", True):
            while reader.isStartElement("Item", False):
                list_.append(reader.readElementContentAsInt("Item"))
            reader.readEndElement("SwitchTable")
        return list_

    @classmethod
    def __loadDirectTable(cls, reader):
        list_ = list()
        if reader.isStartElement("DirectTable", True):
            while reader.isStartElement("Item", True):
                it = GXMacDirectTable()
                list_.append(it)
                it.sourceSId = reader.readElementContentAsInt("SourceSId")
                it.sourceLnId = reader.readElementContentAsInt("SourceLnId")
                it.sourceLcId = reader.readElementContentAsInt("SourceLcId")
                it.destinationSId = reader.readElementContentAsInt("DestinationSId")
                it.destinationLnId = reader.readElementContentAsInt("DestinationLnId")
                it.destinationLcId = reader.readElementContentAsInt("DestinationLcId")
                it.did = GXByteBuffer.hexToBytes(reader.readElementContentAsString("Did"))
            reader.readEndElement("DirectTable")
        return list_

    @classmethod
    def __loadAvailableSwitches(cls, reader):
        list_ = list()
        if reader.isStartElement("AvailableSwitches", True):
            while reader.isStartElement("Item", True):
                it = GXMacAvailableSwitch()
                list_.append(it)
                it.sna = GXByteBuffer.hexToBytes(reader.readElementContentAsString("Sna"))
                it.lsId = reader.readElementContentAsInt("LsId")
                it.level = reader.readElementContentAsInt("Level")
                it.rxLevel = reader.readElementContentAsInt("RxLevel")
                it.rxSnr = reader.readElementContentAsInt("RxSnr")
            reader.readEndElement("AvailableSwitches")
        return list_

    @classmethod
    def __loadCommunications(cls, reader):
        list_ = list()
        if reader.isStartElement("Communications", True):
            while reader.isStartElement("Item", True):
                it = GXMacPhyCommunication()
                list_.append(it)
                it.eui = GXByteBuffer.hexToBytes(reader.readElementContentAsString("Eui"))
                it.txPower = reader.readElementContentAsInt("TxPower")
                it.txCoding = reader.readElementContentAsInt("TxCoding")
                it.rxCoding = reader.readElementContentAsInt("RxCoding")
                it.rxLvl = reader.readElementContentAsInt("RxLvl")
                it.snr = reader.readElementContentAsInt("Snr")
                it.txPowerModified = reader.readElementContentAsInt("TxPowerModified")
                it.txCodingModified = reader.readElementContentAsInt("TxCodingModified")
                it.rxCodingModified = reader.readElementContentAsInt("RxCodingModified")
            reader.readEndElement("Communications")
        return list_

    def load(self, reader):
        self.multicastEntries = self.__loadMulticastEntries(reader)
        self.switchTable = self.__loadSwitchTable(reader)
        self.directTable = self.__loadDirectTable(reader)
        self.availableSwitches = self.__loadAvailableSwitches(reader)
        self.communications = self.__loadCommunications(reader)

    def __saveMulticastEntries(self, writer):
        writer.writeStartElement("MulticastEntries")
        if self.multicastEntries:
            for it in self.multicastEntries:
                writer.writeStartElement("Item")
                writer.writeElementString("Id", it.id)
                writer.writeElementString("Members", it.members)
                writer.writeEndElement()
        writer.writeEndElement()

    def __saveSwitchTable(self, writer):
        writer.writeStartElement("SwitchTable")
        if self.switchTable:
            for it in self.switchTable:
                writer.writeElementString("Item", it)
        writer.writeEndElement()

    def __saveDirectTable(self, writer):
        writer.writeStartElement("DirectTable")
        if self.directTable:
            for it in self.directTable:
                writer.writeStartElement("Item")
                writer.writeElementString("SourceSId", it.sourceSId)
                writer.writeElementString("SourceLnId", it.sourceLnId)
                writer.writeElementString("SourceLcId", it.sourceLcId)
                writer.writeElementString("DestinationSId", it.destinationSId)
                writer.writeElementString("DestinationLnId", it.destinationLnId)
                writer.writeElementString("DestinationLcId", it.destinationLcId)
                writer.writeElementString("Did", GXByteBuffer.hex(it.did, False))
                writer.writeEndElement()
        writer.writeEndElement()

    def __saveAvailableSwitches(self, writer):
        writer.writeStartElement("AvailableSwitches")
        if self.availableSwitches:
            for it in self.availableSwitches:
                writer.writeStartElement("Item")
                writer.writeElementString("Sna", GXByteBuffer.hex(it.sna, False))
                writer.writeElementString("LsId", it.lsId)
                writer.writeElementString("Level", it.level)
                writer.writeElementString("RxLevel", it.rxLevel)
                writer.writeElementString("RxSnr", it.rxSnr)
                writer.writeEndElement()
        writer.writeEndElement()

    def __saveCommunications(self, writer):
        writer.writeStartElement("Communications")
        if self.communications:
            for it in self.communications:
                writer.writeStartElement("Item")
                writer.writeElementString("Eui", GXByteBuffer.hex(it.eui, False))
                writer.writeElementString("TxPower", it.txPower)
                writer.writeElementString("TxCoding", it.txCoding)
                writer.writeElementString("RxCoding", it.rxCoding)
                writer.writeElementString("RxLvl", it.rxLvl)
                writer.writeElementString("Snr", it.snr)
                writer.writeElementString("TxPowerModified", it.txPowerModified)
                writer.writeElementString("TxCodingModified", it.txCodingModified)
                writer.writeElementString("RxCodingModified", it.rxCodingModified)
                writer.writeEndElement()
        writer.writeEndElement()

    def save(self, writer):
        self.__saveMulticastEntries(writer)
        self.__saveSwitchTable(writer)
        self.__saveDirectTable(writer)
        self.__saveAvailableSwitches(writer)
        self.__saveCommunications(writer)



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .GXDLMSObject import GXDLMSObject
from .IGXDLMSBase import IGXDLMSBase
from ..enums import ErrorCode
from ..internal._GXCommon import _GXCommon
from ..GXByteBuffer import GXByteBuffer
from ..enums import ObjectType, DataType
from .enums import MBusEncryptionKeyStatus

# pylint: disable=too-many-instance-attributes
class GXDLMSMBusClient(GXDLMSObject, IGXDLMSBase):
    """
    Online help:
    http://www.gurux.fi/Gurux.DLMS.Objects.GXDLMSMBusClient
    """

    def __init__(self, ln=None, sn=0):
        """
        Constructor.

        ln : Logical Name of the object.
        sn : Short Name of the object.
        """
        #pylint: disable=super-with-arguments
        super(GXDLMSMBusClient, self).__init__(ObjectType.MBUS_CLIENT, ln, sn)
        self.captureDefinition = list()
        self.mBusPortReference = None
        self.capturePeriod = None
        self.primaryAddress = None
        self.identificationNumber = None
        self.manufacturerID = None
        self.dataHeaderVersion = None
        self.deviceType = None
        self.accessNumber = None
        self.status = None
        self.alarm = None
        self.configuration = 0
        self.encryptionKeyStatus = MBusEncryptionKeyStatus.NO_ENCRYPTION_KEY
        self.version = 1

    def slaveInstall(self, client):
        """Installs a slave device."""
        return client.method(self.getName(), self.objectType, 1, 0, DataType.INT8)

    def slaveDeInstall(self, client):
        """De-installs a slave device."""
        return client.method(self.getName(), self.objectType, 2, 0, DataType.INT8)

    def capture(self, client):
        """Captures values."""
        return client.method(self.getName(), self.objectType, 3, 0, DataType.INT8)

    def resetAlarm(self, client):
        """Resets alarm state of the M-Bus slave device."""
        return client.method(self.getName(), self.objectType, 4, 0, DataType.INT8)

    def synchronizeClock(self, client):
        """Synchronize the clock."""
        return client.method(self.getName(), self.objectType, 5, 0, DataType.INT8)

    def sendData(self, client, data):
        """Sends data to the M-Bus slave device."""
        bb = GXByteBuffer()
        bb.setUInt8(DataType.ARRAY)
        bb.setUInt8(DataType.STRUCTURE)
        _GXCommon.setObjectCount(len(data), bb)
        for it in data:
            bb.setUInt8(DataType.STRUCTURE)
            bb.setUInt8(3)
            bb.setUInt8(DataType.OCTET_STRING)
            _GXCommon.setObjectCount(it.DataInformation.Length, bb)
            bb.set(it.DataInformation)
            bb.setUInt8(DataType.OCTET_STRING)
            _GXCommon.setObjectCount(it.ValueInformation.Length, bb)
            bb.set(it.ValueInformation)
            _GXCommon.setData(None, bb, _GXCommon.getDLMSDataType(it.Data), it.Data)
        return client.method(self.getName(), self.objectType, 6, bb, DataType.ARRAY)

    def setEncryptionKey(self, client, encryptionKey):
        """Sets the encryption key in the M-Bus client and enables encrypted communication with the M-Bus slave device."""
        bb = GXByteBuffer()
        bb.setUInt8(DataType.OCTET_STRING)
        if not encryptionKey:
            bb.setUInt8(0)
        else:
            _GXCommon.setObjectCount(encryptionKey.Length, bb)
            bb.set(encryptionKey)
        return client.method(self.getName(), self.objectType, 7, bb, DataType.ARRAY)

    def transferKey(self, client, encryptionKey):
        """Transfers an encryption key to the M-Bus slave device."""
        bb = GXByteBuffer()
        bb.setUInt8(DataType.OCTET_STRING)
        if not encryptionKey:
            bb.setUInt8(0)
        else:
            _GXCommon.setObjectCount(encryptionKey.Length, bb)
            bb.set(encryptionKey)
        return client.method(self.getName(), self.objectType, 8, bb, DataType.ARRAY)

    def getValues(self):
        if self.version == 0:
            return [self.logicalName,
                    self.mBusPortReference,
                    self.captureDefinition,
                    self.capturePeriod,
                    self.primaryAddress,
                    self.identificationNumber,
                    self.manufacturerID,
                    self.dataHeaderVersion,
                    self.deviceType,
                    self.accessNumber,
                    self.status,
                    self.alarm]
        return [self.logicalName,
                self.mBusPortReference,
                self.captureDefinition,
                self.capturePeriod,
                self.primaryAddress,
                self.identificationNumber,
                self.manufacturerID,
                self.dataHeaderVersion,
                self.deviceType,
                self.accessNumber,
                self.status,
                self.alarm,
                self.configuration,
                self.encryptionKeyStatus]

    # Returns collection of attributes to read.  If attribute is static
    #      and
    # already read or device is returned HW error it is not returned.
    def getAttributeIndexToRead(self, all_):
        attributes = list()
        #  LN is static and read only once.
        if all_ or not self.logicalName:
            attributes.append(1)
        #  MBusPortReference
        if all_ or self.canRead(2):
            attributes.append(2)
        #  CaptureDefinition
        if all_ or self.canRead(3):
            attributes.append(3)
        #  CapturePeriod
        if all_ or self.canRead(4):
            attributes.append(4)
        #  PrimaryAddress
        if all_ or self.canRead(5):
            attributes.append(5)
        #  IdentificationNumber
        if all_ or self.canRead(6):
            attributes.append(6)
        #  ManufacturerID
        if all_ or self.canRead(7):
            attributes.append(7)
        #  Version
        if all_ or self.canRead(8):
            attributes.append(8)
        #  DeviceType
        if all_ or self.canRead(9):
            attributes.append(9)
        #  AccessNumber
        if all_ or self.canRead(10):
            attributes.append(10)
        #  Status
        if all_ or self.canRead(11):
            attributes.append(11)
        #  Alarm
        if all_ or self.canRead(12):
            attributes.append(12)
        if self.version != 0:
            #  Alarm
            if all_ or self.canRead(13):
                attributes.append(13)
            #  Alarm
            if all_ or self.canRead(14):
                attributes.append(14)
        return attributes

    #
    # Returns amount of attributes.
    #
    def getAttributeCount(self):
        if self.version != 0:
            return 14
        return 12

    #
    # Returns amount of methods.
    #
    def getMethodCount(self):
        return 8

    def getDataType(self, index):
        if index == 1:
            ret = DataType.OCTET_STRING
        elif index == 2:
            ret = DataType.OCTET_STRING
        elif index == 3:
            ret = DataType.ARRAY
        elif index == 4:
            ret = DataType.UINT32
        elif index == 5:
            ret = DataType.UINT8
        elif index == 6:
            ret = DataType.UINT32
        elif index == 7:
            ret = DataType.UINT16
        elif index == 8:
            ret = DataType.UINT8
        elif index == 9:
            ret = DataType.UINT8
        elif index == 10:
            ret = DataType.UINT8
        elif index == 11:
            ret = DataType.UINT8
        elif index == 12:
            ret = DataType.UINT8
        elif index == 13:
            ret = DataType.UINT16
        elif index == 14:
            ret = DataType.ENUM
        else:
            raise ValueError("getDataType failed. Invalid attribute index.")
        return ret
    #
    # Returns value of given attribute.
    #
    def getValue(self, settings, e):
        #pylint: disable=bad-option-value,redefined-variable-type
        if e.index == 1:
            ret = _GXCommon.logicalNameToBytes(self.logicalName)
        elif e.index == 2:
            ret = _GXCommon.logicalNameToBytes(self.mBusPortReference)
        elif e.index == 3:
            buff = GXByteBuffer()
            buff.setUInt8(DataType.ARRAY)
            _GXCommon.setObjectCount(len(self.captureDefinition), buff)
            for k, v in self.captureDefinition:
                buff.setUInt8(DataType.STRUCTURE)
                buff.setUInt8(2)
                _GXCommon.setData(settings, buff, DataType.UINT8, k)
                if not v:
                    _GXCommon.setData(settings, buff, DataType.OCTET_STRING, None)
                else:
                    _GXCommon.setData(settings, buff, DataType.OCTET_STRING, v.encode())
            ret = buff
        elif e.index == 4:
            ret = self.capturePeriod
        elif e.index == 5:
            ret = self.primaryAddress
        elif e.index == 6:
            ret = self.identificationNumber
        elif e.index == 7:
            ret = self.manufacturerID
        elif e.index == 8:
            ret = self.dataHeaderVersion
        elif e.index == 9:
            ret = self.deviceType
        elif e.index == 10:
            ret = self.accessNumber
        elif e.index == 11:
            ret = self.status
        elif e.index == 12:
            ret = self.alarm
        elif e.index == 13:
            ret = self.configuration
        elif e.index == 14:
            ret = self.encryptionKeyStatus
        else:
            e.error = ErrorCode.READ_WRITE_DENIED
        return ret

    #
    # Set value of given attribute.
    #
    def setValue(self, settings, e):
        if e.index == 1:
            self.logicalName = _GXCommon.toLogicalName(e.value)
        elif e.index == 2:
            self.mBusPortReference = _GXCommon.toLogicalName(e.value)
        elif e.index == 3:
            self.captureDefinition = []
            if e.value:
                for it in e.value:
                    self.captureDefinition.append((_GXCommon.changeType(settings, it[0], DataType.OCTET_STRING), _GXCommon.changeType(settings, it[1], DataType.OCTET_STRING)))
        elif e.index == 4:
            self.capturePeriod = e.value
        elif e.index == 5:
            self.primaryAddress = e.value
        elif e.index == 6:
            self.identificationNumber = e.value
        elif e.index == 7:
            self.manufacturerID = e.value
        elif e.index == 8:
            self.dataHeaderVersion = e.value
        elif e.index == 9:
            self.deviceType = e.value
        elif e.index == 10:
            self.accessNumber = e.value
        elif e.index == 11:
            self.status = e.value
        elif e.index == 12:
            self.alarm = e.value
        elif e.index == 13:
            self.configuration = e.value
        elif e.index == 14:
            self.encryptionKeyStatus = e.value
        else:
            e.error = ErrorCode.READ_WRITE_DENIED

    def load(self, reader):
        self.mBusPortReference = reader.readElementContentAsString("MBusPortReference")
        self.captureDefinition = []
        if reader.isStartElement("CaptureDefinition", True):
            while reader.isStartElement("Item", True):
                d = reader.readElementContentAsString("Data")
                v = reader.readElementContentAsString("Value")
                self.captureDefinition.append((d, v))
            reader.readEndElement("CaptureDefinition")
        self.capturePeriod = reader.readElementContentAsInt("CapturePeriod")
        self.primaryAddress = reader.readElementContentAsInt("PrimaryAddress")
        self.identificationNumber = reader.readElementContentAsInt("IdentificationNumber")
        self.manufacturerID = reader.readElementContentAsInt("ManufacturerID")
        self.dataHeaderVersion = reader.readElementContentAsInt("DataHeaderVersion")
        self.deviceType = reader.readElementContentAsInt("DeviceType")
        self.accessNumber = reader.readElementContentAsInt("AccessNumber")
        self.status = reader.readElementContentAsInt("Status")
        self.alarm = reader.readElementContentAsInt("Alarm")
        if self.version != 0:
            self.configuration = reader.readElementContentAsInt("Configuration")
            self.encryptionKeyStatus = MBusEncryptionKeyStatus(reader.readElementContentAsInt("EncryptionKeyStatus"))

    def save(self, writer):
        writer.writeElementString("MBusPortReference", self.mBusPortReference)
        writer.writeStartElement("CaptureDefinition")
        if self.captureDefinition:
            for k, v in self.captureDefinition:
                writer.writeStartElement("Item")
                writer.writeElementString("Data", k)
                writer.writeElementString("Value", v)
                writer.writeEndElement()
        writer.writeEndElement()
        writer.writeElementString("CapturePeriod", self.capturePeriod)
        writer.writeElementString("PrimaryAddress", self.primaryAddress)
        writer.writeElementString("IdentificationNumber", self.identificationNumber)
        writer.writeElementString("ManufacturerID", self.manufacturerID)
        writer.writeElementString("DataHeaderVersion", self.dataHeaderVersion)
        writer.writeElementString("DeviceType", self.deviceType)
        writer.writeElementString("AccessNumber", self.accessNumber)
        writer.writeElementString("Status", self.status)
        writer.writeElementString("Alarm", self.alarm)
        if self.version != 0:
            writer.writeElementString("Configuration", self.configuration)
            writer.writeElementString("EncryptionKeyStatus", self.encryptionKeyStatus)
